/////////////////////////////////////////////////////////////////////////////
//
// threads.cpp : contains all the different GDI thread classes
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "multithread.h"
#include "threads.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

CRITICAL_SECTION CDemoThread::m_csGDILock;
//	a CRITICAL_SECTION is a segment of code which must be protected during 
//	execution, in this case, while drawing to the device context

HANDLE CDemoThread::m_hAnotherDead = CreateEvent(NULL, FALSE, FALSE, NULL);
//	m_hAnotherDead is used to signal that one or more threads have ended
//	this is an auto-reset event and is intialized as not signaled

/////////////////////////////////////////////////////////////////////////////
//
// CDemoThread : implementation of the CDemoThread class
//
/////////////////////////////////////////////////////////////////////////////

IMPLEMENT_DYNAMIC(CDemoThread, CWinThread)

BEGIN_MESSAGE_MAP(CDemoThread, CWinThread)
	//{{AFX_MSG_MAP(CDemoThread)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

CDemoThread::CDemoThread(CWnd* pWnd, HDC hDC)
{
	m_bAutoDelete = FALSE;
	m_pMainWnd = pWnd;
	m_pMainWnd->GetClientRect( &m_rectClient );
	m_hDC = hDC;

	m_hEventKill = CreateEvent( NULL, TRUE, FALSE, NULL );
	m_hEventDead = CreateEvent( NULL, TRUE, FALSE, NULL );
	//	EventKill and EventDead begin in the signaled state
	//	and must be reset as unsignaled
}

//======================================================================
//	Because MFC GDI objects can not be passed between threads, a CDC 
//	(device context) can not be passd to each thread. Instead, a handle
//	to a device context is passed and the handle is coverted back to 
//	a CDC member within the thread. 
//	If a GDI object were passed to a thread, the object would not be
//	found on the list of GDI objects which MFC maintains for each tread
//	and an ASSERT failure would occur.
//======================================================================

BOOL CDemoThread::InitInstance()
{
	m_dc.Attach(m_hDC);		// thread setup
							// loop but check for kill notification
	while( WaitForSingleObject( m_hEventKill, 0 ) == WAIT_TIMEOUT )
		SingleStep();
	m_dc.Detach();			// thread cleanup
		// return FALSE to avoid entering standard message loop
	return FALSE;
}

void CDemoThread::Delete()
{
		// calling the base here won't do anything but it is a good habit
	CWinThread::Delete();
		// acknowledge receipt of both kill notification events
	VERIFY( SetEvent( m_hEventDead ) );
	VERIFY( SetEvent( m_hAnotherDead ) );
}

CDemoThread::~CDemoThread()
{		// clean up is required 
	CloseHandle( m_hEventKill );
	CloseHandle( m_hEventDead );
}

void CDemoThread::KillThread()
{		// the KillThread function is called in the context of the
		// action threads, not the thread itself
		// resetting the m_hEventKill signals the thread to shutdown
	VERIFY( SetEvent( m_hEventKill ) );
		// during shutdown, the thread is given a higher priority
	SetThreadPriority( THREAD_PRIORITY_ABOVE_NORMAL );
	WaitForSingleObject( m_hEventDead, INFINITE );
	WaitForSingleObject( m_hThread, INFINITE );
		// now delete the CWinThread object
	delete this;
}

void CDemoThread::UpdateBorder()
{		//	Using a CriticalSection really isn't necessary here but
		//	is still good practice. Without the CriticalSection, an 
		//	interrupting thread would receive an intermediate value
		//	but should stabilize later when the new value became valid.
		//	Still, in other circumstances, an abnomality of this nature
		//	could become critical.
	EnterCriticalSection( &CDemoThread::m_csGDILock );
	{
		m_pMainWnd->GetClientRect( &m_rectClient );
	}
	LeaveCriticalSection( &CDemoThread::m_csGDILock );
		//	last, we'll deflate the rectangle slightly (as a margin)
		//	before using it later for tests
	m_rectClient.DeflateRect( 2, 2 );
}


/////////////////////////////////////////////////////////////////////////////
//
// CCircleThread : implementation of the CCircleThread class
//
/////////////////////////////////////////////////////////////////////////////

IMPLEMENT_DYNAMIC(CCircleThread, CDemoThread)

BEGIN_MESSAGE_MAP(CCircleThread, CDemoThread)
	//{{AFX_MSG_MAP(CCircleThread)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

CCircleThread::CCircleThread( CWnd* pWnd,   HDC hDC,     CPoint ptPos, 
						  CPoint ptVel, CSize cSize, COLORREF rgbColor )
						: CDemoThread( pWnd, hDC )
{
	CBrush	cBrush( rgbColor );

	m_hBrush = (HBRUSH) cBrush.Detach();
	m_rectPos[CIRCLE_COUNT].SetRect( ptPos.x, ptPos.y, 
								   ptPos.x + cSize.cx, ptPos.y + cSize.cx );
	m_ptVel	 = ptVel;
	m_nItems = 0;
	m_bErase = FALSE;
}

	//======================================================================
	//	Because MFC GDI objects can not be passed between threads, the
	//	CBrush object can not be passd to each thread. Instead, a handle
	//	to a brush is passed and the handle is coverted back to a CDC
	//	member within the thread. 
	//	If a GDI object were passed to a thread, the object would not be
	//	found on the list of GDI objects which MFC maintains for each
	//	tread and an ASSERT failure would occur.
	//======================================================================

BOOL CCircleThread::InitInstance()
{
	m_cBrush.Attach( m_hBrush );
	return CDemoThread::InitInstance();
}

void CCircleThread::SingleStep()
{
	CRect	cRect;
	CBrush* pOldBrush;
	int		i;

		//	The redraw operation to erase an earlier object is not started
		//	until the array of elements is filled and we're ready to start
		//	discarding older array elements.
	if( m_bErase )
	{
		EnterCriticalSection( &CDemoThread::m_csGDILock );
		{
			pOldBrush = m_dc.SelectObject( &m_cBrush );
			m_dc.SetROP2( R2_XORPEN );
			m_dc.Ellipse( m_rectPos[0] );
			m_dc.SetROP2( R2_COPYPEN );
			m_dc.SelectObject( pOldBrush );
			GdiFlush();
		}
		LeaveCriticalSection( &CDemoThread::m_csGDILock );
	}
	else
		m_bErase = ( m_nItems++ == CIRCLE_COUNT );

		//	Now we need to move all elements down the array
	for( i=0; i<CIRCLE_COUNT; i++ ) m_rectPos[i] = m_rectPos[i+1];
		//	and then update the last element in the array as our new item
	m_rectPos[CIRCLE_COUNT].OffsetRect( m_ptVel );

		//	Since there is no direct test to see if one rectangle falls
		//	inside another (CRect) object, we'll use a union and then test
		//	to see if the resulting rectangle is the same as the smaller of
		//	the two originals. If not, then the smaller rectangle is at least
		//	partially outside the larger.

	cRect.UnionRect( m_rectPos[CIRCLE_COUNT], m_rectClient );
	if( cRect != m_rectClient )
	{
		if( m_rectPos[CIRCLE_COUNT].top < m_rectClient.top 
			&& m_ptVel.y < 0 ) m_ptVel.y *= -1;

		if( m_rectPos[CIRCLE_COUNT].bottom > m_rectClient.bottom 
			&& m_ptVel.y > 0 ) m_ptVel.y *= -1;

		if( m_rectPos[CIRCLE_COUNT].left < m_rectClient.left 
			&& m_ptVel.x < 0 ) m_ptVel.x *= -1;

		if( m_rectPos[CIRCLE_COUNT].right > m_rectClient.right 
			&& m_ptVel.x > 0 ) m_ptVel.x *= -1;
	}
		//	Since all threads share a single device context, a CriticalSection
		//	is used to prevent another thread from usurping access to the GDI
		//	until the drawing operation is completed. 
	EnterCriticalSection(&CDemoThread::m_csGDILock);
	{
		pOldBrush = m_dc.SelectObject( &m_cBrush );
		m_dc.SetROP2( R2_XORPEN );
		m_dc.Ellipse( m_rectPos[CIRCLE_COUNT] );
		m_dc.SetROP2( R2_COPYPEN );
		m_dc.SelectObject( pOldBrush );
			//	To optimize GDI calls, Win32 collects them and thunks batches
			//	of operations on a once per thread basis. Because the GDI is
			//	shared with other threads, the GdiFlush() function is used to 
			//	flush the batch before allowing another thread access.
		GdiFlush();
			//	To experiment, comment out both of the GdiFlush() calls and
			//	observe the results.
	}
	LeaveCriticalSection(&CDemoThread::m_csGDILock);
}


/////////////////////////////////////////////////////////////////////////////
//
// CLineThread : implementation of the CLineThread class
//
/////////////////////////////////////////////////////////////////////////////

IMPLEMENT_DYNAMIC(CLineThread, CDemoThread)

BEGIN_MESSAGE_MAP(CLineThread, CCmdTarget)
	//{{AFX_MSG_MAP(CLineThread)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

CLineThread::CLineThread( CWnd* pWnd, HDC hDC, CPoint ptPos, 
						  CPoint ptVel1, CPoint ptVel2, CSize cSize, 
						  COLORREF rgbColor ) : CDemoThread( pWnd, hDC )
{
	m_ptPos1[LINE_COUNT] = ptPos;
	m_ptPos2[LINE_COUNT] = ptPos + cSize;
	m_ptVel1	 = ptVel1;
	m_ptVel2	 = ptVel2;
	m_nItems	 = 0;
	m_bErase	 = FALSE;

	CPen  cPen( PS_SOLID, 1, rgbColor );
	m_hPen = (HPEN) cPen.Detach();
}

	//======================================================================
	//	Because MFC GDI objects can not be passed between threads, the
	//	CBrush object can not be passd to each thread. Instead, a handle
	//	to a brush is passed and the handle is coverted back to a CDC
	//	member within the thread. 
	//	If a GDI object were passed to a thread, the object would not be
	//	found on the list of GDI objects which MFC maintains for each
	//	tread and an ASSERT failure would occur.
	//======================================================================

BOOL CLineThread::InitInstance()
{
	m_cPen.Attach( m_hPen );
	return CDemoThread::InitInstance();
}

void CLineThread::SingleStep()
{
	CPen*	pOldPen;
	int		i;

		//	The redraw operation to erase an earlier object is not started
		//	until the array of elements is filled and we're ready to start
		//	discarding older array elements.
	if( m_bErase )
	{
		EnterCriticalSection( &CDemoThread::m_csGDILock );
		{
			pOldPen = m_dc.SelectObject( &m_cPen );
			m_dc.SetROP2( R2_XORPEN );
			m_dc.MoveTo( m_ptPos1[0] );
			m_dc.LineTo( m_ptPos2[0] );
			m_dc.SetROP2( R2_COPYPEN );
			m_dc.SelectObject( pOldPen );
			GdiFlush();
		}
		LeaveCriticalSection( &CDemoThread::m_csGDILock );
	}
	else
		m_bErase = ( m_nItems++ == LINE_COUNT );

		//	Now we need to move all elements down the arrays
	for( i=0; i<LINE_COUNT; i++ )
	{
		m_ptPos1[i] = m_ptPos1[i+1];
		m_ptPos2[i] = m_ptPos2[i+1];
	}
		//	and then update the last element in both arrays as our new items
	m_ptPos1[LINE_COUNT].Offset( m_ptVel1 );
	m_ptPos2[LINE_COUNT].Offset( m_ptVel2 );

		//	the PtInRect function offers an easy test to see if the point is
		//	still in the target area or, if not, to run tests to change direction

	if( ! m_rectClient.PtInRect( m_ptPos1[LINE_COUNT] ) )
	{
		if( m_ptPos1[LINE_COUNT].y < m_rectClient.top && m_ptVel1.y < 0 ) 
			m_ptVel1.y *= -1;
		else
		if( m_ptPos1[LINE_COUNT].y > m_rectClient.bottom && m_ptVel1.y > 0 )
			m_ptVel1.y *= -1;

		if( m_ptPos1[LINE_COUNT].x < m_rectClient.left && m_ptVel1.x < 0 ) 
			m_ptVel1.x *= -1;
		else 
		if( m_ptPos1[LINE_COUNT].x > m_rectClient.right && m_ptVel1.x > 0 ) 
			m_ptVel1.x *= -1;
	}
		
	if( ! m_rectClient.PtInRect( m_ptPos2[LINE_COUNT] ) )
	{
		if( m_ptPos2[LINE_COUNT].y < m_rectClient.top && m_ptVel2.y < 0 ) 
			m_ptVel2.y *= -1;
		else
		if( m_ptPos2[LINE_COUNT].y > m_rectClient.bottom && m_ptVel2.y > 0 ) 
			m_ptVel2.y *= -1;

		if( m_ptPos2[LINE_COUNT].x < m_rectClient.left && m_ptVel2.x < 0 ) 
			m_ptVel2.x *= -1;
		else 
		if( m_ptPos2[LINE_COUNT].x > m_rectClient.right && m_ptVel2.x > 0 ) 
			m_ptVel2.x *= -1;
	}
		//	Since all threads share a single device context, a CriticalSection
		//	is used to prevent another thread from usurping access to the GDI
		//	until the drawing operation is completed. 
	EnterCriticalSection(&CDemoThread::m_csGDILock);
	{
		pOldPen = m_dc.SelectObject( &m_cPen );
		m_dc.SetROP2( R2_XORPEN );
		m_dc.MoveTo( m_ptPos1[LINE_COUNT] );
		m_dc.LineTo( m_ptPos2[LINE_COUNT] );
		m_dc.SetROP2( R2_COPYPEN );
		m_dc.SelectObject( pOldPen );
			//	To optimize GDI calls, Win32 collects them and thunks batches
			//	of operations on a once per thread basis. Because the GDI is
			//	shared with other threads, the GdiFlush() function is used to 
			//	flush the batch before allowing another thread access.
		GdiFlush();
			//	To experiment, comment out both of the GdiFlush() calls and
			//	observe the results.
	}
	LeaveCriticalSection( &CDemoThread::m_csGDILock );
}


/////////////////////////////////////////////////////////////////////////////
//
// CRectThread : implementation of the CRectThread class
//
/////////////////////////////////////////////////////////////////////////////

IMPLEMENT_DYNAMIC(CRectThread, CDemoThread)

BEGIN_MESSAGE_MAP(CRectThread, CCmdTarget)
	//{{AFX_MSG_MAP(CRectThread)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

CRectThread::CRectThread( CWnd* pWnd, HDC hDC, CPoint ptPos, CPoint ptVel, 
						  CSize cSize, COLORREF rgbColor) 
						  : CDemoThread( pWnd, hDC )
{
	CBrush	cBrush( rgbColor);

	m_hBrush = (HBRUSH) cBrush.Detach();
	m_rectPos[RECT_COUNT].SetRect( ptPos.x, ptPos.y, 
								   ptPos.x + cSize.cx, ptPos.y + cSize.cy );
	m_ptVel	 = ptVel;
	m_nItems = 0;
	m_bErase = FALSE;
}

	//======================================================================
	//	Because MFC GDI objects can not be passed between threads, the
	//	CBrush object can not be passd to each thread. Instead, a handle
	//	to a brush is passed and the handle is coverted back to a CDC
	//	member within the thread. 
	//	If a GDI object were passed to a thread, the object would not be
	//	found on the list of GDI objects which MFC maintains for each
	//	tread and an ASSERT failure would occur.
	//======================================================================

BOOL CRectThread::InitInstance()
{
	m_cBrush.Attach(m_hBrush);
	return CDemoThread::InitInstance();
}

void CRectThread::SingleStep()
{
	CBrush* pOldBrush;
	CRect	cRect;
	int		i;

		//	The redraw operation to erase an earlier object is not started
		//	until the array of elements is filled and we're ready to start
		//	discarding older array elements.
	if( m_bErase )
	{
		EnterCriticalSection(&CDemoThread::m_csGDILock);
		{
			pOldBrush = m_dc.SelectObject( &m_cBrush );
			m_dc.SetROP2( R2_XORPEN );
			m_dc.Rectangle( m_rectPos[0]);
			m_dc.SetROP2( R2_COPYPEN );
			m_dc.SelectObject( pOldBrush );
			GdiFlush();
		}
		LeaveCriticalSection(&CDemoThread::m_csGDILock);
	}
	else
		m_bErase = ( m_nItems++ == RECT_COUNT );
		//	Now we need to move all elements down the array
	for( i=0; i<RECT_COUNT; i++ ) m_rectPos[i] = m_rectPos[i+1];
		//	and then update the last element in the array as our new item
	m_rectPos[RECT_COUNT].OffsetRect( m_ptVel );

		//	Since there is no direct test to see if one rectangle falls
		//	inside another (CRect) object, we'll use a union and then test
		//	to see if the resulting rectangle is the same as the smaller of
		//	the two originals. If not, then the smaller rectangle is at least
		//	partially outside the larger.
	
	cRect.UnionRect( m_rectPos[RECT_COUNT], m_rectClient );
	if( cRect != m_rectClient )
	{
		if( m_rectPos[RECT_COUNT].top < m_rectClient.top &&
			m_ptVel.y < 0 ) m_ptVel.y *= -1;

		if( m_rectPos[RECT_COUNT].bottom > m_rectClient.bottom && 
			m_ptVel.y > 0 ) m_ptVel.y *= -1;

		if( m_rectPos[RECT_COUNT].left < m_rectClient.left && 
			m_ptVel.x < 0 ) m_ptVel.x *= -1;

		if( m_rectPos[RECT_COUNT].right > m_rectClient.right && 
			m_ptVel.x > 0 ) m_ptVel.x *= -1;
	}
		//	Since all threads share a single device context, a CriticalSection
		//	is used to prevent another thread from usurping access to the GDI
		//	until the drawing operation is completed. 
	EnterCriticalSection( &CDemoThread::m_csGDILock );
	{
		pOldBrush = m_dc.SelectObject( &m_cBrush );
		m_dc.SetROP2( R2_XORPEN );
		m_dc.Rectangle( m_rectPos[RECT_COUNT] );
		m_dc.SetROP2( R2_COPYPEN );
		m_dc.SelectObject( pOldBrush );
			//	To optimize GDI calls, Win32 collects them and thunks batches
			//	of operations on a once per thread basis. Because the GDI is
			//	shared with other threads, the GdiFlush() function is used to 
			//	flush the batch before allowing another thread access.
		GdiFlush();
			//	To experiment, comment out both of the GdiFlush() calls and
			//	observe the results.
	}
	LeaveCriticalSection( &CDemoThread::m_csGDILock );
}
