/*--------------------------------------------------------------------

        PARENT.C  [anonymous pipe version]

        Contains the parent process for the PROCESS demo program.
        In this version, the two processes communicate through
        an anonymous pipe.

        FUNCTIONS
            WinMain             receive and dispatch messages
            Init                initialize program data
            CreateMainWindow    create main overlapping window
            Parent_WndProc      sort out messages for main window
            Parent_OnInitMenu   update status of all menu commands
            Parent_OnCommand    process menu commands
            Parent_OnDestroy    end program
            ChangeShape         store new shape choice
            ChangeSize          store new size choice
            ChangeColor         store new color choice
            StartProcess        start child process
            SendCommand         tell the child process what to draw
            About_DlgProc       sort out messages for About box
            ShowErrorMsg        calls GetLastError and displays the message

        from Mastering Windows NT Programming
        copyright 1993 by Brian Myers & Eric Hamer

  -------------------------------------------------------------------*/

#define STRICT

#include <windows.h>
#include <windowsx.h>
#include <memory.h>             // memset()
#include "parent.h"
#include "global.h"

/*--------------------------------------------------------------------
        GLOBAL VARIABLES
  -------------------------------------------------------------------*/

HWND        hwndMain;           // main overlapping window
HINSTANCE   hInst;              // current instance
HANDLE      hProcess = NULL;    // the child process
HANDLE      hPipeWrite;
FIGURE      figure;             // stores current menu choices

/*------------------------------------------------------------------
        WIN MAIN
        Calls initializing procedures and runs the message loop
  ------------------------------------------------------------------*/

int WINAPI WinMain( HINSTANCE hinstThis, HINSTANCE hinstPrev,
				    LPSTR lpszCmdLine,   int iCmdShow )
{
    MSG msg;

    hInst  = hinstThis;

    // This global FIGURE structure records whatever choices
    // the user makes to choose the shape, size, and color
    // of the figure drawn in the client's window.  Here
    // we initialize it to the program's startup defaults.

    figure.iShape = IDM_RECTANGLE;      // draw a rectangle
    figure.iSize  = IDM_SMALL;          // don't fill the whole window
    figure.iColor = IDM_RED;            // make the rectangle red
    if( ! Init() ) return FALSE;        // register/create window
		// show the main window
    ShowWindow( hwndMain, iCmdShow );
    UpdateWindow( hwndMain );
		// receive and translate messages
    while( GetMessage( &msg, NULL, 0, 0 ) )
    {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
    }
    return( msg.wParam );
    UNREFERENCED_PARAMETER( hinstPrev );
    UNREFERENCED_PARAMETER( lpszCmdLine );
}

/*------------------------------------------------------------------
        INIT
        Register the window classes and the create the windows.
  ------------------------------------------------------------------*/

BOOL Init()
{
    WNDCLASS wc;
    char szAppName[MAX_BUFFER];
    char szTitle[MAX_BUFFER];

    LoadString( hInst, IDS_APPNAME, szAppName,  sizeof(szAppName) );
    LoadString( hInst, IDS_TITLE,   szTitle,    sizeof(szTitle) );

		// register the main window class
    wc.lpszClassName    = szAppName;
    wc.hInstance        = hInst;
    wc.lpfnWndProc      = Parent_WndProc;
    wc.hCursor          = LoadCursor( NULL, IDC_ARROW );
    wc.hIcon            = LoadIcon( hInst, MAKEINTRESOURCE(ICON_APP) );
    wc.lpszMenuName     = MAKEINTRESOURCE( MENU_MAIN );
    wc.hbrBackground    = (HBRUSH)(COLOR_WINDOW + 1);
    wc.style            = CS_HREDRAW | CS_VREDRAW;
    wc.cbClsExtra       = 0;
    wc.cbWndExtra       = 0;
    if( ! RegisterClass( &wc ) )
        return( FALSE );
    return( CreateMainWindow() );
}

/*------------------------------------------------------------------
        CREATE MAIN WINDOW
        Create the program's main overlapping window.
  ------------------------------------------------------------------*/

BOOL CreateMainWindow()
{
    char szAppName[MAX_BUFFER];
    char szTitle[MAX_BUFFER];

		// load the relevant strings
    LoadString( hInst, IDS_APPNAME, szAppName,  sizeof(szAppName));
    LoadString( hInst, IDS_TITLE,   szTitle,    sizeof(szTitle));

		// create the parent window
    hwndMain = CreateWindow( szAppName, szTitle,
							 WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
							 CW_USEDEFAULT, CW_USEDEFAULT, 
							 CW_USEDEFAULT, CW_USEDEFAULT,
							 NULL, NULL, hInst, NULL );
		// return FALSE for an error
    return( hwndMain != NULL );
}

/*------------------------------------------------------------------
        PARENT_WNDPROC
        This is where the messages for the main window are processed.
  ------------------------------------------------------------------*/

LRESULT WINAPI Parent_WndProc( HWND hWnd,	  UINT uMessage,
							   WPARAM wParam, LPARAM lParam )
{
    switch (uMessage)
    {
        HANDLE_MSG( hWnd, WM_INITMENU, Parent_OnInitMenu );
        HANDLE_MSG( hWnd, WM_COMMAND, Parent_OnCommand );
        HANDLE_MSG( hWnd, WM_DESTROY, Parent_OnDestroy );
        default:
            return( DefWindowProc(hWnd, uMessage, wParam, lParam) );
    }
    return( 0L );
}

/*--------------------------------------------------------------------
        PARENT_ONINITMENU
        Check whether the child process exists and enable or disable
        the Start, Terminate, and Options commands accordingly.  Also
        put check marks on the option commands that reflect the
        user's most recent choices.
  --------------------------------------------------------------------*/

void Parent_OnInitMenu( HWND hWnd, HMENU hMenu )
{
    // While the child process does not exist, some of our
    // menu commands make no sense and should be disabled.
    // These include the Terminate command and the figure
    // options commands.

		// get a handle to the options popup menu
    HMENU hmenuOptions = GetSubMenu( hMenu, 1 );

    if (hProcess)
    {		// child process exists; enable Terminate and shape options
        EnableMenuItem( hMenu, IDM_START, MF_GRAYED );
        EnableMenuItem( hMenu, IDM_TERMINATE, MF_ENABLED );
        EnableMenuItem( hmenuOptions, 0, MF_ENABLED | MF_BYPOSITION );
        EnableMenuItem( hmenuOptions, 1, MF_ENABLED | MF_BYPOSITION );
        EnableMenuItem( hmenuOptions, 2, MF_ENABLED | MF_BYPOSITION );
    }
    else
    {		// child process does not exist; disable Terminate and shape options
        EnableMenuItem( hMenu, IDM_START, MF_ENABLED );
        EnableMenuItem( hMenu, IDM_TERMINATE, MF_GRAYED );
        EnableMenuItem( hmenuOptions, 0, MF_GRAYED | MF_BYPOSITION );
        EnableMenuItem( hmenuOptions, 1, MF_GRAYED | MF_BYPOSITION );
        EnableMenuItem( hmenuOptions, 2, MF_GRAYED | MF_BYPOSITION );
    }
		// set a check mark on one of the three shape commands
    CheckMenuItem( hMenu, IDM_ELLIPSE,
				   ( figure.iShape == IDM_ELLIPSE ) ? (int)MF_CHECKED : (int)MF_UNCHECKED );
    CheckMenuItem( hMenu, IDM_RECTANGLE,
				   ( figure.iShape == IDM_RECTANGLE ) ? (int)MF_CHECKED : (int)MF_UNCHECKED );
    CheckMenuItem( hMenu, IDM_TRIANGLE,
				   ( figure.iShape == IDM_TRIANGLE ) ? (int)MF_CHECKED : (int)MF_UNCHECKED );

		// set a check mark on one of the two size commands
    CheckMenuItem( hMenu, IDM_SMALL, 
				   ( figure.iSize == IDM_SMALL ) ? (int)MF_CHECKED : (int)MF_UNCHECKED );
    CheckMenuItem( hMenu, IDM_LARGE, 
				   ( figure.iSize == IDM_LARGE ) ? (int)MF_CHECKED : (int)MF_UNCHECKED );

		// set a check mark on one of the three color commands
    CheckMenuItem( hMenu, IDM_RED, 
				   ( figure.iColor == IDM_RED ) ? (int)MF_CHECKED : (int)MF_UNCHECKED );
    CheckMenuItem( hMenu, IDM_GREEN, 
				   ( figure.iColor == IDM_GREEN ) ? (int)MF_CHECKED : (int)MF_UNCHECKED );
    CheckMenuItem( hMenu, IDM_BLUE, 
				   ( figure.iColor == IDM_BLUE ) ? (int)MF_CHECKED : (int)MF_UNCHECKED );


    return;
    UNREFERENCED_PARAMETER( hWnd );
}

/*------------------------------------------------------------------
        PARENT_ONCOMMAND
        Process commands from the menu.
  ------------------------------------------------------------------*/

void Parent_OnCommand( HWND hWnd, int iCmd, HWND hWndCtl, UINT uCode )
{
    switch (iCmd)
    {
        case IDM_START:			// start the drawing process
            StartProcess();
            break;

        case IDM_ELLIPSE:		// send the figure data through the pipe
        case IDM_RECTANGLE:
        case IDM_TRIANGLE:
        case IDM_TERMINATE:		// when child sees TERMINATE it quits
            ChangeShape( iCmd );
            break;

        case IDM_SMALL:			// adjust the size of the figure
        case IDM_LARGE:
            ChangeSize( iCmd );
            break;

        case IDM_RED:			// change the color of the figure
        case IDM_BLUE:
        case IDM_GREEN:
            ChangeColor( iCmd );
            break;

        case IDM_ABOUT:			// show the about dialog box
            DialogBox( hInst, MAKEINTRESOURCE(DLG_ABOUT),
                       hWnd, About_DlgProc );
            break;

        case IDM_EXIT:			// exit from the program 
            DestroyWindow( hWnd );	// destroy program window
            break;

        default:
            break;
    }
    return;
    UNREFERENCED_PARAMETER( hWndCtl );
    UNREFERENCED_PARAMETER( uCode );
}

/*------------------------------------------------------------------
        PARENT_ONDESTROY
        Post the quit message.
  ------------------------------------------------------------------*/

void Parent_OnDestroy ( HWND hWnd )
{
    PostQuitMessage(0);
    return;
    UNREFERENCED_PARAMETER( hWnd );
}

/*------------------------------------------------------------------
        CHANGE SHAPE
        Record a shape command from the user.  If the user has chosen
        a new shape, send the updated FIGURE structure to the child.
  ------------------------------------------------------------------*/

void ChangeShape ( int iCmd )
{
    if( iCmd != figure.iShape )		// new shape?
    {
        figure.iShape = iCmd;		// record it
        SendCommand();				// tell child process
    }
    return;
}

/*------------------------------------------------------------------
        CHANGE SIZE
        Record a size command from the user.  If the user has chosen
        a new size, send the updated FIGURE structure to the child.
  ------------------------------------------------------------------*/

void ChangeSize ( int iCmd )
{
    if( iCmd != figure.iSize )		// new size?
    {
        figure.iSize = iCmd;		// record it
        SendCommand();				// tell child process
    }
    return;
}

/*------------------------------------------------------------------
        CHANGE COLOR
        Record a color command from the user.  If the user has chosen
        a new color, send the updated FIGURE structure to the child.
  ------------------------------------------------------------------*/

void ChangeColor ( int iCmd )
{
    if( iCmd != figure.iColor )		// new color?
    {
        figure.iColor = iCmd;		// record it
        SendCommand();				// tell child process
    }
    return;
}

/*------------------------------------------------------------------
        START PROCESS
        In response to the IDM_START command, launch the child
        process and create the pipe for talking to it.
  ------------------------------------------------------------------*/

void StartProcess ( void )
{
    SECURITY_ATTRIBUTES sa;			// security privileges for handles
    STARTUPINFO			sui;		// info for starting a process
    PROCESS_INFORMATION pi;			// info returned about a process
    char	szProcess[MAX_BUFFER];	// name of child process image
    int		iLen;					// return value
    BOOL	bTest;					// return value
    HANDLE	hPipeRead;				// inbound end of pipe (for client)
    
		// load name of process image file from resources 
    iLen = LoadString( hInst, IDS_PROCESS, szProcess, sizeof(szProcess) );
    if( ! iLen ) return;
		// fill out a SECURITY_ATTRIBUTES structure so handles can be inherited
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);   // structure size
    sa.lpSecurityDescriptor = NULL;             // default descriptor
    sa.bInheritHandle = TRUE;                   // inheritable
		// create the pipe 
    bTest = CreatePipe( &hPipeRead,				// reading handle
						&hPipeWrite,            // writing handle
						&sa,                    // lets handles be inherited
						0 );                    // default buffer size
    if( ! bTest )
    {
        ShowErrorMsg();
        return;                         // error during pipe creation
    }
		// make an uninheritable duplicate of the outbound (write) handle
    bTest = DuplicateHandle( GetCurrentProcess(),
						     hPipeWrite,				// original handle
							 GetCurrentProcess(),
							 NULL,						// don't create new handle
							 0,
							 FALSE,						// not inheritable
							 DUPLICATE_SAME_ACCESS );

    if( ! bTest )
    {
        ShowErrorMsg();
        CloseHandle( hPipeRead );
        CloseHandle( hPipeWrite );
        return;                             // duplication failed
    }
		// fill in the process's startup information
    memset( &sui, 0, sizeof(STARTUPINFO) );
    sui.cb         = sizeof(STARTUPINFO);
    sui.dwFlags    = STARTF_USESTDHANDLES;
    sui.hStdInput  = hPipeRead;
    sui.hStdOutput = GetStdHandle( STD_OUTPUT_HANDLE );
    sui.hStdError  = GetStdHandle( STD_ERROR_HANDLE );
		// create the drawing process
    bTest = CreateProcess( szProcess,		// .EXE image
						   NULL,			// command line
						   NULL,			// process security
						   NULL,			// thread security
						   TRUE,			// inherit handles--yes
						   0,				// creation flags
						   NULL,			// environment block
						   NULL,			// current directory
						   &sui,			// startup info
						   &pi );			// process info (returned)
		// did we succeed in launching the process?
    if( ! bTest )
    {
        ShowErrorMsg( );					// creation failed
        CloseHandle( hPipeWrite );
    }
    else									// creation succeeded
    {
        hProcess = pi.hProcess;				// save new process handle
        CloseHandle( pi.hThread );			// discard new thread handle
        figure.iShape = IDM_RECTANGLE;		// reset to default shape
        SendCommand();						// tell child what to draw
    }

    CloseHandle( hPipeRead );				// discard receiving end of pipe
    return;
}

/*--------------------------------------------------------------------
        SEND COMMAND
        Tell the child program what to draw.  Write the current
        contents of the global FIGURE variable into the pipe.

        Return
        TRUE indicates the write operation succeeded.  FALSE means
        an error occurred and we have lost contact with the child.
  --------------------------------------------------------------------*/

BOOL SendCommand()
{
    BOOL bTest;                     // return value
    DWORD dwWritten;                // number of bytes written to pipe

    /* pass the choices to the child through the pipe */
    bTest = WriteFile( hPipeWrite,		// anonymous pipe (outbound handle)
					   &figure,         // buffer to write
					   sizeof(FIGURE),  // size of buffer
					   &dwWritten,      // bytes written
					   NULL );          // overlapping i/o structure

    if( ! bTest )						// did writing succeed?
    {
        // If the write operation failed because the user has
        // already closed the child program, then tell the user
        // the connection was broken.  If some less predictable
        // error caused the failure, call ShowErrorMsg as usual
        // to display the system's error message.

        DWORD dwResult = GetLastError();

        if( dwResult == ERROR_BROKEN_PIPE ||     // pipe has been ended
            dwResult == ERROR_NO_DATA )		     // pipe close in progress
        {
				// Presumably the user closed the child.
            MessageBox( hwndMain, "Connection with child was already broken.",
						"Parent Message", MB_ICONEXCLAMATION | MB_OK );
        }
        else    // an unpredictable error occurred
        {
            ShowErrorMsg( );
        }
    }

    // If a write error occurred, or if we just sent an IDM_TERMINATE
    // command to make the child quit, then in either case we break
    // off communication with the child process.

    if ((!bTest) || (figure.iShape == IDM_TERMINATE))
    {
        CloseHandle( hProcess );        // forget about the child
        hProcess = NULL;
        CloseHandle( hPipeWrite );      // destroy the pipe
    }

    return( bTest );
}

/*------------------------------------------------------------------
        ABOUT BOX DIALOG PROCEDURE
        Process messages for the About box window here.
  ------------------------------------------------------------------*/

BOOL WINAPI About_DlgProc( HWND hDlg,		UINT uMessage,
						   WPARAM wParam,	LPARAM lParam )
{
    switch( uMessage )
    {
        case WM_INITDIALOG:								return( TRUE );
        case WM_COMMAND:	EndDialog( hDlg, TRUE );	return( TRUE );
    }

    return( FALSE );
    UNREFERENCED_PARAMETER( wParam );
    UNREFERENCED_PARAMETER( lParam );
}

/*--------------------------------------------------------------------
        SHOW ERROR MESSAGE
  --------------------------------------------------------------------*/

void ShowErrorMsg ( void )
{
    LPVOID lpvMessage;                      // temporary message buffer

		// retrieve a message from the system message table
    FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
				   NULL, GetLastError(),
			       MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),
				   (LPTSTR) &lpvMessage, 0, NULL );
		// display the message in a message box
    MessageBox( hwndMain, lpvMessage, "Parent Message",
                MB_ICONEXCLAMATION | MB_OK );
		// release the buffer FormatMessage allocated
    LocalFree( lpvMessage );
    return;
}
