/*--------------------------------------------------------------------

	SENDMAIL.CPP
   
	Demonstrates using Win32 Winsock API functions to perform email
   transmission using the SMTP protocol. Takes four parameters, as 
   follows:
   
      sendmail mailserv to_addr from_addr messagefile.txt
      
   Connects to the specified mail server, uses to_addr and from_addr
   in the transmitted message header, then sends each line of the 
   messagefile (text file) as the email message body. Transfer is
   implemented using the SMTP Internet protocol.
   
   WINSOCK FUNCTIONS USED:

      WSAStartup              Initializes specific winsock version.
      gethostbyname           Attempts to lookup host name from IP #.
      getservbyname           Looks up service port from service name.
      socket                  Creates a socket of given type.
      htons                   Converts a USHORT from host byte order. 
      connect                 Establishes connection on a socket.
      recv                    Receives data through socket.
      send                    Transmits data through socket.
      closesocket             Closes (deallocates) a socket.
      WSACleanup              Deallocates internal winsock resources.
   
   NOTES: 
    - This program supports both Windows 95 and NT 4.0, and requires 
      Microsoft TCP/IP networking (or another TCP/IP stack) to be 
      installed and configured prior to use. (You will also need to
      have a TCP/IP connection established or else have auto-dialing
      enabled).

		written/updated by Jim Blaney, 1998
		for Windows 98 Developers Handbook

  -------------------------------------------------------------------*/

#define WIN32_LEAN_AND_MEAN
#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>
#include <iostream.h>
#include <windows.h>
#include <winsock2.h>

// Insist on at least Winsock v1.1, can be changed to 2.0
const    VERSION_MAJOR  = 1;     
const    VERSION_MINOR  = 1;

#define  CRLF  "\r\n"   // carriage-return/line feed pair.


void ShowUsage(void)
{
   cout << "Usage: SENDMAIL mailserv to_addr from_addr messagefile" << endl;
   cout << "Example: SENDMAIL smtp.myisp.com rcvr@elsewhere.com ";
   cout << "my_id@mydomain.com message.txt" << endl;
   exit(1);
}

// Basic error checking for send() and recv() functions.
void Check( int iStatus, char *szFunction )
{
   if (iStatus != SOCKET_ERROR && iStatus != 0) return;

   cerr << "Error during call to " << szFunction << ": ";
   cerr << iStatus << " - " << GetLastError() << endl;
}


int main(int argc, char *argv[])
{
   WSADATA        WSData;
   LPHOSTENT      lpHostEntry;
   LPSERVENT      lpServEntry;
   SOCKADDR_IN    SockAddr;
   SOCKET         hServer;
   int            iProtocolPort;
   char           szSmtpServerName[64], szToAddr[64], szFromAddr[64];
   char           szBuffer[4096], szLine[255], szMsgLine[255];
   
   // Check for four command-line args.
   if (argc != 5) ShowUsage();
   
   // Load command-line args (for clarity).
   lstrcpy( szSmtpServerName, argv[1]);
   lstrcpy( szToAddr,         argv[2]);
   lstrcpy( szFromAddr,       argv[3]);

   // Create input stream for reading email message file.
   ifstream MsgFile( argv[4] );

   // Attempt to intialize WinSock (1.1 or later).
   if ( WSAStartup(MAKEWORD(VERSION_MAJOR, VERSION_MINOR), &WSData) ) 
   {
      cout << "Cannot find Winsock v";
      cout << VERSION_MAJOR << "." << VERSION_MINOR;
      cout << " or later!" << endl;
      return(1);
   }

   // Lookup email server's IP address.
   lpHostEntry = gethostbyname( szSmtpServerName );
   if (lpHostEntry == NULL) 
   {
      cout << "Cannot find SMTP mail server ";
      cout << szSmtpServerName << endl;
      return(1);
   }
   
   // Create a TCP/IP socket, no specific protocol.
   hServer = socket( PF_INET, SOCK_STREAM, 0); 
   if (hServer == INVALID_SOCKET) 
   {
      cout << "Cannot open mail server socket\n";
      return(1);
   }

   // Get the mail service port. 
   lpServEntry = getservbyname( "mail", 0);

   // Use the SMTP default port if no other port is specified.
   if (lpServEntry == NULL)
     iProtocolPort = htons(IPPORT_SMTP);
   else
     iProtocolPort = lpServEntry->s_port;

   // Setup a Socket Address structure.
   SockAddr.sin_family = AF_INET;
   SockAddr.sin_port = iProtocolPort;
   SockAddr.sin_addr = *((LPIN_ADDR)*lpHostEntry->h_addr_list);

   // Connect the Socket.
   if (connect( hServer, (PSOCKADDR) &SockAddr, sizeof(SockAddr)))
   {
     cout << "Error connecting to Server socket.\n";
     return (1);
   }

   // Receive initial response from SMTP server.
   Check( recv( hServer, szBuffer, sizeof(szBuffer), 0), "recv() Reply");

   // Send HELO server.com
   sprintf( szMsgLine, "HELO %s%s", szSmtpServerName, CRLF);
   Check( send( hServer, szMsgLine, strlen(szMsgLine), 0), "send() HELO");
   Check( recv( hServer, szBuffer, sizeof(szBuffer), 0), "recv() HELO");

   // Send MAIL FROM: <sender@mydomain.com>
   sprintf( szMsgLine, "MAIL FROM:<%s>%s", szFromAddr, CRLF);
   Check( send( hServer, szMsgLine, strlen(szMsgLine), 0), "send() MAIL FROM");
   Check( recv( hServer, szBuffer, sizeof(szBuffer), 0), "recv() MAIL FROM");

   // Send RCPT TO: <receiver@domain.com>
   sprintf( szMsgLine, "RCPT TO:<%s>%s", szToAddr, CRLF);
   Check( send( hServer, szMsgLine, strlen(szMsgLine), 0), "send() RCPT TO");
   Check( recv( hServer, szBuffer, sizeof(szBuffer), 0), "recv() RCPT TO");

   // Send DATA.
   sprintf( szMsgLine, "DATA%s", CRLF);
   Check( send( hServer, szMsgLine, strlen(szMsgLine), 0), "send() DATA");
   Check( recv( hServer, szBuffer, sizeof(szBuffer), 0), "recv() DATA");

   // Send all lines of message body (using supplied text file.)
   MsgFile.getline( szLine, sizeof(szLine)); // get first line.
   do // for each line of message text...
   {
      //if (!szLine[0]) lstrcpy(szLine, " "); // avoid sending 0 bytes for a line.
      sprintf( szMsgLine, "%s%s", szLine, CRLF);
      Check( send( hServer, szMsgLine, strlen(szMsgLine), 0), "send() message-line");
      MsgFile.getline( szLine, sizeof(szLine)); // get next line.
   } while ( MsgFile.good() );

   // Send blank line and a period.
   sprintf( szMsgLine, "%s.%s", CRLF, CRLF);
   Check( send( hServer, szMsgLine, strlen(szMsgLine), 0), "send() end-message");
   Check( recv( hServer, szBuffer, sizeof(szBuffer), 0), "recv() end-message");

   // Send QUIT.
   sprintf( szMsgLine, "QUIT%s", CRLF);
   Check( send( hServer, szMsgLine, strlen(szMsgLine), 0), "send() QUIT");
   Check( recv( hServer, szBuffer, sizeof(szBuffer), 0), "recv() QUIT");

   // Report message has been sent.
   cout << "Sent " << argv[4] << " as email message to " << szToAddr << endl;
   
   // Close server socket and prepare to exit.
   closesocket( hServer );
   WSACleanup();

   return(0);
}
 
  
