/*--------------------------------------------------------------------

	RPCSERV.CPP

	Demonstrates using RPC in a server application to make calls on 
   behalf of an RPC client application. Functions callable by the 
   client are defined in the mathlib.cpp source file.
   
   FUNCTIONS USED:

   RpcServerUseProtseqEp   Specifies the RPC protocol and endpoint.
   RpcServerRegisterIf     Registers an interface with RPC.
   RpcServerInqBindings    Returns binding handles used for RPC.
   RpcNsBindingExport      Establishes name in name-service database.
   RpcServerListen         Listens for RPC calls.
   RpcMgmtWaitServerListen Performs specified RPC wait operation.

   midl_user_allocate      Memory allocation function used by RPC.
   midl_user_free          Memory freeing function used by RPC.

   Additionally, the following functions are called by this server 
   (via the RPC mechanism), in response to calls by RPCLIENT.EXE:

   Add                  Simple math functions which get called via RPC.
   Subtract              "      "     "        "     "     "    "   "
   Multiply              "      "     "        "     "     "    "   "
   Divide                "      "     "        "     "     "    "   "
   SendClientName       Demo's passing a string to server app.
   GetMathServerName    Demo's retrieving a string from server app.
   
   NOTES: 
   - This program will run on Windows 98, depending on the transport
   used.
   
		written/updated by Jim Blaney, 1998
		for Windows 98 Developers Handbook

  -------------------------------------------------------------------*/
#include <iostream.h>
#include <rpc.h>
#include "..\rpcmathlib\mathlib.h"

void _CRTAPI1 main(int argc, char *argv[])
{
   RPC_STATUS           status;
   RPC_BINDING_VECTOR   *pBindingVector;

   UCHAR *pszProtocolSequence = (UCHAR *) "ncacn_np";   
   WORD  wMinCalls            = 1;
   WORD  wMaxCalls            = 30;
   UCHAR *pszEndPoint         = (UCHAR *) "\\pipe\\auto";   
   WORD  fWaitFlag            = FALSE;
   UCHAR *pszEntryName        = (UCHAR *) "/.:/Autohandle_mathlib";

   status = RpcServerUseProtseqEp(  
      pszProtocolSequence,    // protocol sequence
      wMaxCalls,              // maximum concurrent calls
      pszEndPoint,            // end point (using pipe)
      NULL );                 // security (none used)
       
   if (status) 
   {
      cout << "error "<< status << " setting endpoint!\n";
      exit;
   }

   status = RpcServerRegisterIf(
      mathlib_v1_0_s_ifspec,  // Interface
      NULL,                   // Default MgrTypeUUID 
      NULL  );                // Default MgrEpv

   if (status) 
   {
      cout << "error "<< status << " registering server!\n";
      exit;
   }
   

   status = RpcServerInqBindings( &pBindingVector );

   if (status) 
   {
      cout << "error "<< status << " ?? bindings!\n";
      exit;
   }

   status = RpcNsBindingExport(
      RPC_C_NS_SYNTAX_DEFAULT,   // name syntax type
      pszEntryName,              // Name space name
      mathlib_v1_0_s_ifspec,     // interface handle
      pBindingVector,            // Binding vector (we received)
      NULL );                    // UUID vector

   if (status) 
   {
      cout << "error "<< status << " Exporting bindings!\n";
      exit;
   }


   cout << "Listening for RPC calls..." << endl << flush;

   status = RpcServerListen(
      1,                         // minimum calls
      30,                        // maximum calls
      FALSE );                   // wait value (FALSE = DO wait)

   if (status) 
   {
      cout << "error "<< status << " during listen!\n";
      exit;
   }

               
   if ( fWaitFlag )
   {
      status = RpcMgmtWaitServerListen(); // Wait for RPC calls.

      if (status) 
      {
         cout << "error "<< status << " during listen!\n";
         exit;
      }
   }
}


// -----------------------------------------------------------------------
// The following are two memory allocation/deallocation functions which 
// get called via the RPC mechanism, in order to allocate and free memory 
// used for pointers, strings, and other function parameters which use 
// additional memory.
// -----------------------------------------------------------------------
void __RPC_FAR *__RPC_API midl_user_allocate( size_t len )
{
   return( malloc(len) );
}

void __RPC_API midl_user_free( void __RPC_FAR *ptr )
{
   free( ptr );
}
 
