/*--------------------------------------------------------------------

        GRAPHWIN.C

        Routines to handle the Graph window from the SHOWWAVE
        program.  The Graph window is the control in the center
        of the program's modeless dialog box where a graph of
        the current sound wave appears.

        PUBLIC FUNCTIONS
            Graph_WndProc       window procedure for graph control
            RegisterGraphClass  register graph control window class

        PRIVATE FUNCTION
            Graph_OnPaint       paint wave line in graph control

        program by Michael Barnes
        copyright 1992, 1993 by Brian Myers, Chris Doner, & Eric Hamer
        from Mastering Windows NT Programming

  --------------------------------------------------------------------*/

#define STRICT
#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>           // multimedia definitions
#include "showwave.h"           // program variables and functions

/*--------------------------------------------------------------------
        FUNCTION PROTOTYPES
  --------------------------------------------------------------------*/

LRESULT WINAPI Graph_WndProc( HWND hWnd, UINT uMsg, WPARAM wParam,
    LPARAM lParam );
static void Graph_OnPaint( HWND hWnd );

/*--------------------------------------------------------------------
        STATIC VARIABLES
  --------------------------------------------------------------------*/

static char szGraphClass[] = "GraphClass";

/*--------------------------------------------------------------------
        MACROS
  --------------------------------------------------------------------*/

#define RECTWIDTH( r )  ((r.right) - (r.left) + 1)
#define RECTHEIGHT( r ) ((r.bottom) - (r.top) + 1)

/*--------------------------------------------------------------------
        REGISTER GRAPH CLASS
        Register the window class for the dialog box's wave graph
        control window.  The main dialog box's resource template
        names this window class for one of its controls.  This
        procedure must be called before CreateDialog.
  --------------------------------------------------------------------*/

BOOL RegisterGraphClass ( HINSTANCE hInstance )
{
    WNDCLASS wc;

    wc.style         = 0;
    wc.lpfnWndProc   = Graph_WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = NULL;
    wc.hCursor       = LoadCursor( NULL, IDC_ARROW );
    wc.hbrBackground = GetStockBrush( BLACK_BRUSH );
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = szGraphClass;

    return( RegisterClass(&wc) );
}

/*--------------------------------------------------------------------
        GRAPH WNDPROC
        Receive messages for the main dialog box's sound graph window.
  --------------------------------------------------------------------*/

LRESULT WINAPI Graph_WndProc (
    HWND hWnd,
    UINT uMsg,
    WPARAM wParam,
    LPARAM lParam )
{
    switch (uMsg)
    {
        /* paint the green sound graph line */
        HANDLE_MSG( hWnd, WM_PAINT, Graph_OnPaint );

        default:
            return( DefWindowProc(hWnd, uMsg, wParam, lParam) );
    }
}

/*--------------------------------------------------------------------
        GRAPH_ONPAINT
        Handle WM_PAINT messages for the Graph window.  Draw
        the green wave graph.
  --------------------------------------------------------------------*/

static void Graph_OnPaint ( HWND hWnd )
{
    PAINTSTRUCT ps;
    HDC   hDC;
    HPEN  hPen, hpenOld;            // green pen for drawing columns
    RECT  rClient;                  // size of graph control window
    int   iBase;                    // vertical position of baseline
    LPSTR lp;                       // points to one sample in wave data
    DWORD dwMaxStart;               // maximum value for starting point
    int   iYScale;                  // vertical scaling factor
    int   iCol;                     // horizontal position of a column
    int   iColHeight;               // height of a column

    /* begin paint processing */
    hDC = BeginPaint( hWnd, &ps );

    /* create a pen for drawing graph */
    if (GetDeviceCaps(hDC, NUMCOLORS) > 2)
    {
        hPen = CreatePen( PS_SOLID, 1, RGB_GREEN );     // color display
    }
    else
    {
        hPen = CreatePen( PS_SOLID, 1, RGB_WHITE );     // mono display
    }

    if (hPen)
    {
        /* select the pen */
        hpenOld = SelectPen( hDC, hPen );

        /* draw the waveform baseline */
        GetClientRect( hWnd, &rClient );
        iBase = RECTHEIGHT(rClient) / 2;
        MoveToEx( hDC, 0, iBase, NULL );
        LineTo( hDC, (int)rClient.right, iBase );

        /* graph waveform data */
        if (lpWaveData)
        {
            /* set the current sample position in the waveform data */
            dwMaxStart = dwWaveDataSize - RECTWIDTH(rClient) - 1;
            lp = lpWaveData + min( dwCurrentSample, dwMaxStart );

            /* determine the height scaling factor */
            iYScale = ((MAXVALUE-MINVALUE) + 1) /   // amplitude
                      (RECTHEIGHT(rClient) - 4);    // control height

                      // Subtracting 4 from the height ensures a small
                      // margin above and below the biggest waves

            /* paint samples from the waveform data */
            for (iCol = (int)rClient.left;
                iCol <= (int)rClient.right;
                iCol++)
            {
                iColHeight = ((int)(BYTE)(*lp++) - MIDVALUE) / iYScale;
                if (iColHeight != 0)
                {
                    /* figure absolute value of column height */
                    if (iColHeight < 0)
                    {
                         iColHeight = -iColHeight;
                    }

                    /* draw line from below base to above base */
                    MoveToEx( hDC, iCol, iBase-iColHeight, NULL );
                    LineTo( hDC, iCol, iBase + (iColHeight+1) );
                }
            }
        }

        /* restore the DC and delete the pen */
        SelectPen( hDC, hpenOld );
        DeletePen( hPen );
    }

    /* end paint processing */
    EndPaint( hWnd, &ps );
    return;
}
