/*--------------------------------------------------------------------

        MCI.C
        Routines for the SHOWWAVE program to manipulate audio using
        the Multimedia Command Interface (MCI).

        PUBLIC FUNCTIONS
            OpenDevice          open audio device
            CloseDevice         close audio device
            SetTimeFormat       choose millisecond time format
            BeginPlay           begin sound playback
            StopPlay            end sound playback
            BeginRecord         begin recording
            SaveRecord          save recording

        PRIVATE FUNCTION
            ReportMCIError      show error message for user

        program by Michael Barnes
        copyright 1992, 1993 by Brian Myers, Chris Doner, & Eric Hamer
        from Mastering Windows NT Programming

  --------------------------------------------------------------------*/

#define STRICT
#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>           // multimedia definitions
#include "rsrc.h"               // program resource ID definitions

/*--------------------------------------------------------------------
        FUNCTION PROTOTYPE
  --------------------------------------------------------------------*/

static void ReportMCIError( HWND hWnd, DWORD dwError );

/*--------------------------------------------------------------------
        REPORT MCI ERROR
        Report given MCI error to the user
  --------------------------------------------------------------------*/

static void ReportMCIError (
    HWND hWnd,
    DWORD dwError )
{
    HINSTANCE hInstance;
    char szErrStr[MAXERRORLENGTH];
    char szCaption[MAX_RSRC_STRING_LEN];

    hInstance = GetWindowInstance( hWnd );
    LoadString( hInstance, IDS_CAPTION, szCaption, sizeof(szCaption) );
    mciGetErrorString( dwError, szErrStr, sizeof(szErrStr) );
    MessageBox( hWnd, szErrStr, szCaption, MB_ICONEXCLAMATION | MB_OK );

    return;
}

/*--------------------------------------------------------------------
        OPEN DEVICE
        Open a waveaudio device
  --------------------------------------------------------------------*/

BOOL OpenDevice (
    HWND hWnd,
    LPSTR lpszFileName,
    MCIDEVICEID *lpmciDevice )
{
    DWORD dwRet;
    MCI_OPEN_PARMS mciOpenParms;

    /* open the compound device */
    mciOpenParms.lpstrDeviceType  = "waveaudio";
    mciOpenParms.lpstrElementName = lpszFileName;
    dwRet = mciSendCommand( 0,                      // device id
                MCI_OPEN,                           // command
                MCI_OPEN_TYPE | MCI_OPEN_ELEMENT,   // flags
                (DWORD)(LPVOID)&mciOpenParms );     // param block

    if (dwRet != 0)
    {
        ReportMCIError( hWnd, dwRet );
        return( FALSE );
    }

    /* set return values */
    *lpmciDevice = mciOpenParms.wDeviceID;
    return( TRUE );
}

/*--------------------------------------------------------------------
        CLOSE DEVICE
        Close a multimedia device
  --------------------------------------------------------------------*/

void CloseDevice (
    HWND hWnd,
    MCIDEVICEID mciDevice )
{
    DWORD dwRet;

    dwRet = mciSendCommand( mciDevice,
                            MCI_CLOSE,
                            MCI_WAIT,
                            (DWORD)NULL );

    if (dwRet != 0)
    {
        ReportMCIError( hWnd, dwRet );
    }
    return;
}

/*--------------------------------------------------------------------
        SET TIME FORMAT
        Set time format.  Use milliseconds (not bytes or samples).
  --------------------------------------------------------------------*/

BOOL SetTimeFormat (
    HWND hWnd,
    MCIDEVICEID mciDevice )
{
    DWORD dwRet;
    MCI_SET_PARMS mciSetParms;

    /* set time format to milliseconds */
    mciSetParms.dwTimeFormat = MCI_FORMAT_MILLISECONDS;
    dwRet = mciSendCommand( mciDevice,
                            MCI_SET,
                            MCI_SET_TIME_FORMAT,
                            (DWORD)(LPVOID)&mciSetParms );

    if (dwRet != 0)
    {
        ReportMCIError( hWnd, dwRet );
        return( FALSE );
    }

    return( TRUE );                         // success
}

/*--------------------------------------------------------------------
        BEGIN PLAYBACK
  --------------------------------------------------------------------*/

BOOL BeginPlay (
    HWND hWnd,
    MCIDEVICEID mciDevice,
    DWORD dwFrom )
{
    DWORD dwRet;
    MCI_PLAY_PARMS mciPlayParms;

    /* set time format to milliseconds */
    if (!SetTimeFormat( hWnd, mciDevice ))
    {
        return( FALSE );
    }

    // The callback window will be notified with an MM_MCINOTIFY message
    // when playback is complete.  At that time, the window procedure
    // closes the device.

    mciPlayParms.dwCallback = (DWORD)(LPVOID)hWnd;
    mciPlayParms.dwFrom = dwFrom;
    dwRet = mciSendCommand( mciDevice,
                            MCI_PLAY,
                            MCI_FROM | MCI_NOTIFY,
                            (DWORD)(LPVOID)&mciPlayParms );

    if (dwRet != 0)
    {
        ReportMCIError( hWnd, dwRet );
        return( FALSE );
    }

    return( TRUE );                         // success
}

/*--------------------------------------------------------------------
        STOP PLAY
        Terminate playback
  --------------------------------------------------------------------*/

void StopPlay (
    HWND hWnd,
    MCIDEVICEID mciDevice )
{
    DWORD dwRet;

    dwRet = mciSendCommand( mciDevice,
                            MCI_STOP,
                            MCI_WAIT,
                            (DWORD)NULL );

    if (dwRet != 0)
    {
        ReportMCIError( hWnd, dwRet );
    }
    return;
}

/*--------------------------------------------------------------------
        BEGIN RECORD
  --------------------------------------------------------------------*/

BOOL BeginRecord (
    HWND hWnd,
    MCIDEVICEID mciDevice,
    DWORD dwTo )
{
    DWORD dwRet;
    MCI_RECORD_PARMS mciRecordParms;

    /* set time format to milliseconds */
    if (!SetTimeFormat( hWnd, mciDevice ))
    {
        return( FALSE );
    }

    // Begin recording for the specified number of milliseconds.
    // The callback window will be notified with an MM_MCINOTIFY message
    // when recording is complete.  At that time, the window procedure
    // saves the recording and closes the device.

    mciRecordParms.dwCallback = (DWORD)(LPVOID) hWnd;
    mciRecordParms.dwTo = dwTo;
    dwRet = mciSendCommand( mciDevice,
                            MCI_RECORD,
                            MCI_TO | MCI_NOTIFY,
                            (DWORD)(LPVOID)&mciRecordParms );

    if (dwRet != 0)
    {
        ReportMCIError( hWnd, dwRet );
        return( FALSE );
    }

    return( TRUE );                         // success
}

/*--------------------------------------------------------------------
        SAVE RECORD
        Save recording
  --------------------------------------------------------------------*/

BOOL SaveRecord (
    HWND hWnd,
    MCIDEVICEID mciDevice,
    LPSTR lpszFileName )
{
    DWORD dwRet;
    MCI_SAVE_PARMS mciSaveParms;

    // Save the recording to the specified file.  Wait for
    // the operation to complete before continuing.

    mciSaveParms.lpfilename = lpszFileName;
    dwRet = mciSendCommand( mciDevice,
                            MCI_SAVE,
                            MCI_SAVE_FILE | MCI_WAIT,
                            (DWORD)(LPVOID)&mciSaveParms );

    if (dwRet != 0)
    {
        ReportMCIError( hWnd, dwRet );
        return( FALSE );
    }

    return( TRUE );                         // success
}
