/*--------------------------------------------------------------------

        SHOWWAVE.C

        Routines for handling the SHOWWAVE window, the modeless dialog
        that serves as the main window for the ShowWave program.
        Primarily these routines maintain the program window and
        respond to commands from the user.

        PUBLIC FUNCTIONS
            ShowWave_WndProc        callback function for main window

        MESSAGE HANDLING FUNCTIONS
            ShowWave_OnInitDialog   initialize ShowWave modeless dialog box
            ShowWave_OnCtlColor     set dialog box's background color
            ShowWave_OnCommand      handle menu commands
            ShowWave_OnHScroll      handle waveform data scrolling
            ShowWave_OnTimer        handle timer request to update display
            ShowWave_OnMCINotify    handle MM_MCINOTIFY message
            ShowWave_OnQueryEndSession handle request to end demo window session
            ShowWave_OnDestroy      handle window destruction

        OTHER PRIVATE FUNCTIONS
            PaintWaveData           paint the waveform data
            UpdateFileTitleText     update file title text
            UpdateTimeText          update Position and Length text
            ConfirmClose            confirm closing a dirty file with user
            GetFileName             invoke the file name common dialog boxes
            FreeGlobalWaveData      free global waveform data
            CloseAudioDevice        close an opened waveform audio device
            CloseTimerDevice        kill an active timer
            NewWaveFile             start work on a new .WAV file
            OpenWaveFile            open a new .WAV file
            SaveWaveFile            save waveform data to disk
            MixWithFile             mix another .WAV file into current file
            ChangeVolume            change the playback volume
            PlayWaveFile            play waveform data
            RecordWaveFile          record waveform data

        program by Michael Barnes
        copyright 1992, 1993 by Brian Myers, Chris Doner, & Eric Hamer
        from Mastering Windows NT Programming

  --------------------------------------------------------------------*/

#define STRICT
#include <windows.h>
#include <windowsx.h>
#include <commdlg.h>            // common dialog box (for file names)
#include <mmsystem.h>           // multimedia definitions
#include <stdlib.h>             // _MAX_* constants
#include "showwave.h"           // program variables and functions
#include "rsrc.h"               // program resource ID definitions

/*--------------------------------------------------------------------
        PROTOTYPES FOR PRIVATE FUNCTIONS
  --------------------------------------------------------------------*/

static BOOL ShowWave_OnInitDialog( HWND hDlg, HWND hWndFocus, LPARAM lParam );
static HBRUSH ShowWave_OnCtlColor( HWND hDlg, HDC hDC, HWND hCtl, int iType );
static void ShowWave_OnCommand( HWND hDlg, int iCmd, HWND hCtl, UINT uCode );
static void ShowWave_OnHScroll( HWND hDlg, HWND hCtl, UINT uCode, int iPos );
static void ShowWave_OnTimer( HWND hDlg, UINT uID );
static void ShowWave_OnMCINotify( HWND hDlg, UINT uCode, UINT uDevID );
static BOOL ShowWave_OnQueryEndSession( HWND hDlg );
static void ShowWave_OnDestroy( HWND hDlg );

static void PaintWaveData( HWND hDlg );
static void UpdateFileTitleText( HWND hDlg );
static void UpdateTimeText( HWND hDlg );
static BOOL GetFileName( HWND hDlg, BOOL bOpenName, LPSTR lpszFile,
    int iMaxFileNmLen, LPSTR lpszFileTitle, int iMaxFileTitleLen );
static void FreeGlobalWaveData( void );
static void CloseAudioDevice( HWND hDlg );
static void CloseTimerDevice( HWND hDlg );
static void NewWaveFile( HWND hDlg );
static void OpenWaveFile( HWND hDlg );
static void SaveWaveFile( HWND hDlg, BOOL bAskForName );
static void MixWithFile( HWND hDlg );
static void ChangeVolume( HWND hDlg, BOOL bIncrease );
static void PlayWaveFile( HWND hDlg );
static void RecordWaveFile( HWND hDlg );

/*--------------------------------------------------------------------
        STATIC VARIABLES
  --------------------------------------------------------------------*/

static char szOFNDefExt[]     = "WAV";  // for file name dialog
static char *szOFNFilter[]    =
{
    "Sound Files (*.WAV)\0 *.WAV\0, All Files(*.*)\0    *.* \0\0"
};

static int iHScrollPos = 0;             // scroll bar thumb position 

static char szFileName[_MAX_PATH];      // file handling 
static char szFileTitle[_MAX_FNAME];
static BOOL bIsFileDirty = FALSE;

static UINT uTimerID  = 0;              // MCI symbols and variables 
static MCIDEVICEID mciDevice = 0;
static char szTempFileName[_MAX_PATH];  // stores new unnamed sounds
#define TMRPLAY         1               // timer event IDs 
#define TMRRECORD       2

static HBRUSH hbrBackgnd;               // for dialog box's background 

/*--------------------------------------------------------------------
        SHOWWAVE DIALOG PROCEDURE
        Sort out messages for the program's main window, the ShowWave
        dialog box.
  --------------------------------------------------------------------*/

LRESULT WINAPI ShowWave_WndProc (
    HWND hDlg,
    UINT uMsg,
    WPARAM wParam,
    LPARAM lParam )
{
    switch (uMsg)
    {
        HANDLE_MSG( hDlg, WM_INITDIALOG, ShowWave_OnInitDialog );

        HANDLE_MSG( hDlg, WM_CTLCOLORSTATIC, ShowWave_OnCtlColor );

        HANDLE_MSG( hDlg, WM_CTLCOLORBTN, ShowWave_OnCtlColor );

        HANDLE_MSG( hDlg, WM_COMMAND, ShowWave_OnCommand );

        HANDLE_MSG( hDlg, WM_HSCROLL, ShowWave_OnHScroll );

        HANDLE_MSG( hDlg, WM_TIMER, ShowWave_OnTimer );

        HANDLE_MSG( hDlg, MM_MCINOTIFY, ShowWave_OnMCINotify );

        HANDLE_MSG( hDlg, WM_QUERYENDSESSION, ShowWave_OnQueryEndSession );

        HANDLE_MSG( hDlg, WM_DESTROY, ShowWave_OnDestroy );

        default:
            return( DefWindowProc(hDlg, uMsg, wParam, lParam) );
     }
}

/*--------------------------------------------------------------------
        SHOWWAVE_ONINITDIALOG
        Handle WM_INITDIALOG here. Initialize program settings.
  --------------------------------------------------------------------*/

static BOOL ShowWave_OnInitDialog (
    HWND hDlg,
    HWND hwndFocus,
    LPARAM lParam )
{
    char szTempPath[_MAX_PATH];
    
    /* create a brush for painting the dialog box's background */
    hbrBackgnd = CreateSolidBrush( GetSysColor(COLOR_BTNFACE) );

    /* get file name for temporary storage of waveform data */
    GetTempPath( sizeof(szTempPath), szTempPath );
    GetTempFileName( szTempPath, "SWT", 0, szTempFileName );

    /* start out with an empty wave file */
    FORWARD_WM_COMMAND( hDlg, IDM_NEW, NULL, 0, PostMessage );

    /* do nothing with the focus */
    return( TRUE );
    UNREFERENCED_PARAMETER( hwndFocus );
    UNREFERENCED_PARAMETER( lParam );
}

/*--------------------------------------------------------------------
        SHOWWAVE_ONCTLCOLOR
        Process WM_CTLCOLOR messages here.  Give text controls
        a white background.  Shade other parts of dialog.

        RETURN
        A brush handle to use in painting the background.  The
        handle used is declared with the other static variables
        near the top of this file.
  --------------------------------------------------------------------*/

static HBRUSH ShowWave_OnCtlColor (
    HWND hDlg,                          // parent dialog window
    HDC hDC,                            // DC for control's client area
    HWND hCtl,                          // control window
    int iType )                         // type of control
{
	switch (iType)
	{
		case CTLCOLOR_STATIC:
            /* colors for static text controls */
			SetTextColor( hDC, GetSysColor(COLOR_BTNTEXT) );
			SetBkColor( hDC, RGB_WHITE );
			return( GetStockBrush(WHITE_BRUSH) );

		default:
				/* any other controls */
			SetTextColor( hDC, GetSysColor(COLOR_BTNTEXT) );
			SetBkColor( hDC, GetSysColor(COLOR_BTNFACE) );
			return( hbrBackgnd );
	}

//	return;
	UNREFERENCED_PARAMETER( hDlg );
	UNREFERENCED_PARAMETER( hCtl );
}

/*--------------------------------------------------------------------
        SHOWWAVE_ONCOMMAND
        Handle WM_COMMAND messages here.  Respond to actions from
        each dialog control and from the menu.
  --------------------------------------------------------------------*/

static void ShowWave_OnCommand (
    HWND hDlg,
    int iCmd,
    HWND hCtl,
    UINT uCode )
{
    switch (iCmd)
    {
        case IDM_NEW:                       // clear data buffer
            NewWaveFile( hDlg );
            break;

        case IDM_OPEN:                      // load a disk file
            OpenWaveFile( hDlg );
            break;

        case IDM_SAVE:                      // save to disk file
        case IDM_SAVEAS:
            SaveWaveFile( hDlg, (iCmd==IDM_SAVEAS) );
            break;

        case IDM_EXIT:                      // end program
            FORWARD_WM_CLOSE( hDlg, PostMessage );
            break;

        case IDM_MIXWITHFILE:               // mix two sounds
            MixWithFile( hDlg );
            break;

        case IDM_INCREASEVOLUME:            // make louder
        case IDM_DECREASEVOLUME:            // make softer
            ChangeVolume( hDlg, iCmd==IDM_INCREASEVOLUME );
            break;

        case IDM_ABOUT:                     // show About box
            DialogBox( GetWindowInstance(hDlg),
                MAKEINTRESOURCE(DLG_ABOUT),
                hDlg, About_DlgProc );
            break;

        case IDD_PLAY:                      // play a sound
            PlayWaveFile( hDlg );
            break;

        case IDD_RECORD:                    // record a sound
            RecordWaveFile( hDlg );
            break;

        case IDD_STOP:                      // interrupt device
            if (mciDevice)
            {
                StopPlay( hDlg, mciDevice );
            }
            break;

        default:
            break;
    }
    
    return;
    UNREFERENCED_PARAMETER( hCtl );
    UNREFERENCED_PARAMETER( uCode );
}

/*--------------------------------------------------------------------
        SHOWWAVE_ONHSCROLL
        Process WM_HSCROLL messages here.  Advance through the
        sound file according to the user's signals from the scroll
        bar.
  --------------------------------------------------------------------*/

static void ShowWave_OnHScroll (
    HWND hDlg,
    HWND hCtl,
    UINT uCode,
    int iPos )
{
    int iMinPos, iMaxPos;

    ScrollBar_GetRange( hCtl, &iMinPos, &iMaxPos );

    switch (uCode)
    {
        case SB_LINEUP:
            iHScrollPos -= 11;
            break;

        case SB_LINEDOWN:
            iHScrollPos += 11;
            break;

        case SB_PAGEUP:
            iHScrollPos -= 101;
            break;

        case SB_PAGEDOWN:
            iHScrollPos += 101;
            break;

        case SB_TOP:
            iHScrollPos = iMinPos;
            break;

        case SB_BOTTOM:
            iHScrollPos = iMaxPos;
            break;

        case SB_THUMBPOSITION:
        case SB_THUMBTRACK:
            iHScrollPos = iPos;
            break;

        default:
            return;
    }

    /* update scroll bar thumb */
    iHScrollPos = max( iMinPos, min(iHScrollPos, iMaxPos) );
    ScrollBar_SetPos( hCtl, iHScrollPos, TRUE );

    /* set current sample */
    dwCurrentSample = (iHScrollPos*dwSamplesPerSec) / 100;

    /* set position and length text */
    UpdateTimeText( hDlg );

    /* paint the waveform data */
    PaintWaveData( hDlg );
    return;
}

/*--------------------------------------------------------------------
        SHOWWAVE_ONTIMER
        Handle WM_TIMER messages here.  Update display to show
        current position in sound file.
  --------------------------------------------------------------------*/

void ShowWave_OnTimer (
    HWND hDlg,
    UINT uID )
{
    int  iMaxPos;
    HWND hwndScrollBar = GetDlgItem( hDlg, IDD_SCROLLBAR );

    if (uID == TMRPLAY)
    {
        /* set the new scroll bar position */
        FORWARD_WM_HSCROLL( hDlg, hwndScrollBar, 0, SB_LINEDOWN,
                            SendMessage );
    }
    else
    {
        /* set the new waveform data */
        dwWaveDataSize += dwSamplesPerSec/10;

        /* set the new scroll bar range */
        iMaxPos = (int)((dwWaveDataSize*100) / dwSamplesPerSec );
        ScrollBar_SetRange( hwndScrollBar, 0, iMaxPos, FALSE );

        /* set the new scroll bar position */
        FORWARD_WM_HSCROLL( hDlg, hwndScrollBar, 0, SB_BOTTOM,
                            SendMessage );
    }
    return;
}

/*--------------------------------------------------------------------
        SHOWWAVE_ONMCINOTIFY
        Handle MM_MCINOTIFY messages here.  Clean up after a playback
        or recording operation terminates.
  --------------------------------------------------------------------*/

static void ShowWave_OnMCINotify (
    HWND hDlg,
    UINT uCode,
    MCIDEVICEID mciDevice )
{
    int  iMaxPos;
    HWND hwndScrollBar = GetDlgItem( hDlg, IDD_SCROLLBAR );

    if (uTimerID == TMRPLAY)
    {
        /* close devices */
        CloseAudioDevice( hDlg );
        CloseTimerDevice( hDlg );
        if (uCode == MCI_NOTIFY_SUCCESSFUL)
        {
            FORWARD_WM_HSCROLL( hDlg, hwndScrollBar, 0, SB_BOTTOM,
                                SendMessage );
        }
    }
    else
    {
        /* save recording and close devices */
        SaveRecord( hDlg, mciDevice, szTempFileName );
        CloseAudioDevice( hDlg );
        CloseTimerDevice( hDlg );

        /* set file dirty flag */
        bIsFileDirty = TRUE;

        /* read new waveform data */
        ReadWaveData( hDlg, szTempFileName,
                      &lpWaveData,
                      &dwWaveDataSize,
                      &dwSamplesPerSec );

        /* set the new scroll bar range */
        iMaxPos = (int)((dwWaveDataSize*100) / dwSamplesPerSec);
        ScrollBar_SetRange( hwndScrollBar, 0, iMaxPos, FALSE );

        /* set the new scroll bar position */
        FORWARD_WM_HSCROLL( hDlg, hwndScrollBar, 0, SB_BOTTOM,
                            SendMessage );
    }

    return;
    UNREFERENCED_PARAMETER( mciDevice );
}

/*--------------------------------------------------------------------
        SHOWWAVE_ONQUERYENDSESSION
        Handle WM_QUERYENDSESSION messages here.  Give the user a
        chance to save any open file before program ends.
  --------------------------------------------------------------------*/

static BOOL ShowWave_OnQueryEndSession ( HWND hDlg )
{
    return( QuerySave(hDlg) );
}

/*--------------------------------------------------------------------
        SHOWWAVE_ONDESTROY
        Handle WM_DESTROY message here.  Close all devices.
  --------------------------------------------------------------------*/

static void ShowWave_OnDestroy ( HWND hDlg )
{
    OFSTRUCT of;

    DeleteBrush( hbrBackgnd );
    FreeGlobalWaveData( );                          // release buffer
    CloseAudioDevice( hDlg );                       // close devices
    CloseTimerDevice( hDlg );
    OpenFile( szTempFileName, &of, OF_DELETE );     // delete temp file

    PostQuitMessage( 0 );                           // send WM_QUIT
    return;
}

/*--------------------------------------------------------------------
        PAINT WAVE DATA
        Repaint the dialog box's GraphClass display control
  --------------------------------------------------------------------*/

static void PaintWaveData ( HWND hDlg )
{
    HWND hwndShowWave = GetDlgItem( hDlg, IDD_SHOWWAVE );

    InvalidateRect( hwndShowWave, NULL, TRUE );
    UpdateWindow( hwndShowWave );
    return;
}

/*--------------------------------------------------------------------
        UPDATE FILE TITLE TEXT
        Set a new file name in the dialog box's static file name control
  --------------------------------------------------------------------*/

static void UpdateFileTitleText ( HWND hDlg )
{
    Static_SetText( GetDlgItem(hDlg, IDD_FILETITLE), szFileTitle );
    return;
}

/*--------------------------------------------------------------------
        UPDATE TIME TEXT
        Update the static dialog controls that show the scroll
        thumb's current position and the playing time for the
        current .WAV file.
  --------------------------------------------------------------------*/

static void UpdateTimeText ( HWND hDlg )
{
    DWORD dwFrac;
    UINT  uSecs, uTenthSecs;
    char  szText[MAX_RSRC_STRING_LEN];
    char  szFormat[MAX_RSRC_STRING_LEN];

    /* get the format string */
    LoadString( GetWindowInstance(hDlg), IDS_TIMEFMT, szFormat,
                sizeof(szFormat) );

    /* update position text */
    dwFrac     = ((dwCurrentSample*100) / dwSamplesPerSec) / 10;
    uSecs      = (UINT)(dwFrac / 10);
    uTenthSecs = (UINT)(dwFrac % 10);
    wsprintf( szText, szFormat, uSecs, uTenthSecs );
    Static_SetText( GetDlgItem(hDlg, IDD_POSITION), szText );

    /* update length text */
    dwFrac     = ((dwWaveDataSize*100) / dwSamplesPerSec) / 10;
    uSecs      = (UINT)(dwFrac / 10);
    uTenthSecs = (UINT)(dwFrac % 10);
    wsprintf( szText, szFormat, uSecs, uTenthSecs );
    Static_SetText( GetDlgItem(hDlg, IDD_LENGTH), szText );

    return;
}

/*--------------------------------------------------------------------
        QUERY SAVE
        Ask user to confirm loss of unsaved file before closing
  --------------------------------------------------------------------*/

BOOL QuerySave ( HWND hDlg )
{
    HINSTANCE hInstance;
    char szText[MAX_RSRC_STRING_LEN];
    char szCaption[MAX_RSRC_STRING_LEN];
    char szFormat[MAX_RSRC_STRING_LEN];
    int  iRet;

    /* is file dirty? */
    if (!bIsFileDirty)
    {
        return( TRUE );
    }

    /* see if user wants to save the modifications */
    hInstance = GetWindowInstance( hDlg );
    LoadString( hInstance, IDS_CAPTION,      szCaption, sizeof(szCaption) );
    LoadString( hInstance, IDS_CONFIRMCLOSE, szFormat,  sizeof(szFormat)  );
    wsprintf( szText, szFormat, (LPSTR)szFileTitle );

    iRet = MessageBox( hDlg, szText, szCaption,
                       MB_YESNOCANCEL | MB_ICONQUESTION );
    if (iRet == IDYES)
    {
        FORWARD_WM_COMMAND( hDlg, IDM_SAVE, NULL, 0, SendMessage );
    }
    return( iRet != IDCANCEL );
}

/*--------------------------------------------------------------------
        GET FILE NAME
        Invoke the File Open or File Save As common dialog box.
        If the bOpenName parameter is TRUE, the procedure runs
        the OpenFileName dialog box.

        RETURN
        TRUE if the dialog box closes without error.  If the dialog
        box returns TRUE, then lpszFile and lpszFileTitle point to
        the new file path and name, respectively.
  --------------------------------------------------------------------*/

static BOOL GetFileName (
    HWND hDlg,
    BOOL bOpenName,                     // open file or save as
    LPSTR lpszFile,                     // buffer for file path
    int iMaxFileNmLen,                  // max file path length
    LPSTR lpszFileTitle,                // buffer for file name
    int iMaxFileTitleLen )              // max file name length
{
    OPENFILENAME ofn;

    /* initialize structure for the common dialog box */
    lpszFile[0] = '\0';
    ofn.lStructSize       = sizeof( OPENFILENAME );
    ofn.hwndOwner         = hDlg;
    ofn.hInstance         = NULL;
    ofn.lpstrFilter       = szOFNFilter[0];
    ofn.lpstrCustomFilter = NULL;
    ofn.nMaxCustFilter    = 0;
    ofn.nFilterIndex      = 1;
    ofn.lpstrFile         = lpszFile;
    ofn.nMaxFile          = iMaxFileNmLen;
    ofn.lpstrFileTitle    = lpszFileTitle;
    ofn.nMaxFileTitle     = iMaxFileTitleLen;
    ofn.lpstrInitialDir   = NULL;
    ofn.lpstrTitle        = NULL;
    ofn.nFileOffset       = 0;
    ofn.nFileExtension    = 0;
    ofn.lpstrDefExt       = szOFNDefExt;
    ofn.lCustData         = 0;
    ofn.lpfnHook          = NULL;
    ofn.lpTemplateName    = NULL;

    /* invoke the common dialog box */
    if (bOpenName)                              // Open a file
    {
        ofn.Flags = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST |
                    OFN_FILEMUSTEXIST;
        return( GetOpenFileName(&ofn) );
    }
    else                                        // Save As...
    {
        ofn.Flags = OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT;
        return( GetSaveFileName(&ofn) );
    }
}

/*--------------------------------------------------------------------
        FREE GLOBAL WAVE DATA
        Free storage associated with the global waveform data
  --------------------------------------------------------------------*/

static void FreeGlobalWaveData ( void )
{
    if (lpWaveData)
    {
        GlobalFreePtr( lpWaveData );
        lpWaveData = NULL;
    }
    return;
}

/*--------------------------------------------------------------------
        CLOSE AUDIO DEVICE
        Close an opened waveform audio device
  --------------------------------------------------------------------*/

static void CloseAudioDevice ( HWND hDlg )
{
    if (mciDevice)
    {
        CloseDevice( hDlg, mciDevice );
        mciDevice = 0;
    }
    return;
}

/*--------------------------------------------------------------------
        CLOSE TIMER DEVICE
        Kill an active timer; stop receiving WM_TIMER messages
  --------------------------------------------------------------------*/

static void CloseTimerDevice ( HWND hDlg )
{
    if (uTimerID)
    {
        KillTimer( hDlg, uTimerID );
        uTimerID = 0;
    }
    return;
}

/*--------------------------------------------------------------------
        NEW WAVE FILE
        Start work on a new .WAV file.  Reset variables and update
        display.  Called in response to the New command.
  --------------------------------------------------------------------*/

static void NewWaveFile ( HWND hDlg )
{
    HINSTANCE hInstance;
    HWND hwndScrollBar = GetDlgItem( hDlg, IDD_SCROLLBAR );

     /* close the old Wave file */
    if (!QuerySave( hDlg ))
    {
        return;
    }

    /* set file name and title */
    hInstance = GetWindowInstance( hDlg );
    LoadString( hInstance, IDS_UNTITLED, szFileTitle,
                sizeof(szFileTitle) );
    szFileName[0] = '\0';
    FreeGlobalWaveData( );                  // delete old waveform data
    bIsFileDirty = FALSE;                   // set file dirty flag
    UpdateFileTitleText( hDlg );            // set file name text

    /* set the new waveform data */
    dwCurrentSample = 0;
    lpWaveData      = NULL;
    dwWaveDataSize  = 0;
    dwSamplesPerSec = 22050;

    /* set the new scroll bar range */
    ScrollBar_SetRange( hwndScrollBar, 0, 1, FALSE );

    /* set the new scroll bar position */
    FORWARD_WM_HSCROLL( hDlg, hwndScrollBar, 0, SB_TOP, SendMessage );

    return;
}

/*--------------------------------------------------------------------
        OPEN WAVE FILE
        Open a new .WAV file
  --------------------------------------------------------------------*/

static void OpenWaveFile ( HWND hDlg )
{
    int  iMaxPos;
    HWND hwndScrollBar = GetDlgItem( hDlg, IDD_SCROLLBAR );

    /* close the old Wave file */
    if (!QuerySave( hDlg ))
    {
        return;
    }

    /* get a Wave file to open */
    if (!GetFileName( hDlg, TRUE,
                      szFileName,
                      sizeof(szFileName),
                      szFileTitle,
                      sizeof(szFileTitle) ))
    {
        return;
    }

    FreeGlobalWaveData( );                  // delete old waveform data
    bIsFileDirty = FALSE;                   // set file dirty flag
    UpdateFileTitleText( hDlg );            // set file name text

    /* read new waveform data */
    dwCurrentSample = 0;
    ReadWaveData( hDlg,
                  szFileName,
                  &lpWaveData,
                  &dwWaveDataSize,
                  &dwSamplesPerSec );

    /* set the new scroll bar range */
    iMaxPos = (int)((dwWaveDataSize*100) / dwSamplesPerSec );
    ScrollBar_SetRange( hwndScrollBar, 0, iMaxPos, FALSE );

    /* set the new scroll bar position */
    FORWARD_WM_HSCROLL( hDlg, hwndScrollBar, 0, SB_TOP, SendMessage );

    return;
}

/*--------------------------------------------------------------------
        SAVE WAVE FILE
        Save waveform data to disk.  If the second parameter is TRUE,
        or if the current data does not yet have a file name, this
        procedure will request a name.
  --------------------------------------------------------------------*/

static void SaveWaveFile (
    HWND hDlg,
    BOOL bAskForName )
{
    HINSTANCE hInstance;
    BOOL bSave;
    char szText[MAX_RSRC_STRING_LEN];
    char szCaption[MAX_RSRC_STRING_LEN];
    int  iRet;

    /* anything to save? */
    if (!lpWaveData)
    {
        return;
    }

    /* if renaming or no name, prompt user for name */
    if ((bAskForName) || (szFileName[0] == '\0'))
    {
        /* get the name of the wavefile to save as */
        bSave = GetFileName( hDlg, FALSE,
                             szFileName,  sizeof(szFileName),
                             szFileTitle, sizeof(szFileTitle) );
    }
    else
    {
        /* no new name; confirm overwriting old file */
        hInstance = GetWindowInstance( hDlg );
        LoadString( hInstance, IDS_OVERWRITE, szText,
                    sizeof(szText) );
        LoadString( hInstance, IDS_CAPTION, szCaption,
                    sizeof(szCaption) );
        iRet = MessageBox( hDlg, szText, szCaption,
                    MB_YESNO | MB_ICONQUESTION );
        bSave = (iRet == IDYES);
    }

    /* save to the Wave file */
    if (bSave)
    {
        /* set file dirty flag */
        bIsFileDirty = FALSE;

        /* set file name text */
        UpdateFileTitleText( hDlg );

        /* write out the waveform data */
        WriteWaveData( hDlg, szFileName,
                       lpWaveData,
                       dwWaveDataSize,
                       dwSamplesPerSec );
    }
    return;
}

/*--------------------------------------------------------------------
        MIX WITH FILE
        Mix a .WAV file into the current waveform data, combining
        the data from both into a single recording.  Mixing begins
        at the current point in the current file.  Mixing stops if
        we reach the end of the current file; it will not extend
        the current file.
  --------------------------------------------------------------------*/

static void MixWithFile ( HWND hDlg )
{
    HINSTANCE hInstance;
    char  szMixFile[_MAX_FNAME];            // name of second .WAV file
    LPSTR lpMix;                            // data from second file
    DWORD dwMixSize;                        // size of new data
    DWORD dwMixSPS;                         // samples per second
    char  szErrStr[MAX_RSRC_STRING_LEN];
    char  szCaption[MAX_RSRC_STRING_LEN];
    LPSTR lpDest, lpSrc;                    // pointers for data transfer
    DWORD dw;                               // loop variable
    int   iMaxPos;                          // scroll bar range
    HWND  hwndScrollBar = GetDlgItem( hDlg, IDD_SCROLLBAR );

    /* get a .WAV file to mix with */
    if (!GetFileName( hDlg, TRUE,
                      szMixFile,
                      sizeof(szMixFile),
                      NULL, 0 ))
    {
        return;                             // no file name
    }

    /* read waveform data */
    if (!ReadWaveData( hDlg,
                       szMixFile,
                       &lpMix,
                       &dwMixSize,
                       &dwMixSPS ))
    {
        return;                             // error reading data
    }

    /* mix data */
    if (!lpWaveData)
    {
        /* set file dirty flag */
        bIsFileDirty = TRUE;

        /* set the new waveform data */
        dwCurrentSample = 0;
        lpWaveData      = lpMix;
        dwWaveDataSize  = dwMixSize;
        dwSamplesPerSec = dwMixSPS;

        /* set the new scroll bar range */
        iMaxPos = (int)((dwWaveDataSize*100) / dwSamplesPerSec );
        ScrollBar_SetRange( hwndScrollBar, 0, iMaxPos, FALSE );

        /* set the new scroll bar position */
        FORWARD_WM_HSCROLL( hDlg, hwndScrollBar, 0, SB_TOP, SendMessage );
    }
    else
    {
        /* for demo, use only matching frequencies */
        if (dwSamplesPerSec != dwMixSPS)
        {
            hInstance = GetWindowInstance( hDlg );
            LoadString( hInstance, IDS_BADFREQUENCY, szErrStr,
                        sizeof(szErrStr) );
            LoadString( hInstance, IDS_CAPTION, szCaption,
                        sizeof(szCaption) );
            MessageBox( hDlg, szErrStr, szCaption,
                        MB_ICONEXCLAMATION | MB_OK );
            GlobalFreePtr( lpMix );
            return;
        }

        /* mix new file into waveform at current position (lpDest) */
        lpSrc  = lpMix;
        lpDest = lpWaveData + dwCurrentSample;
        for (dw = 0; dw < (dwWaveDataSize-dwCurrentSample); dw++)
        {
            /* merge one source and destination sample */
            *lpDest = (BYTE)(((int)(BYTE)*lpDest + (BYTE)*lpSrc) / 2);

            lpSrc++;                // increment transfer pointers
            lpDest++;

            if (lpSrc >= (lpMix+dwMixSize))
            {
                break;              // reached end of original data
            }
        }

        /* clean up */
        GlobalFreePtr( lpMix );     // free memory
        bIsFileDirty = TRUE;        // set file dirty flag
        PaintWaveData( hDlg );      // paint the new waveform data
    }
    return;
}

/*--------------------------------------------------------------------
        CHANGE VOLUME
        Increase/decrease the volume of the waveform data playback
  --------------------------------------------------------------------*/

static void ChangeVolume (
    HWND hDlg,
    BOOL bIncrease )
{
    LPSTR lp;
    DWORD dw;
    int   iTmp, iFactor;

    /* anything to change? */
    if (!lpWaveData)
    {
        return;
    }

    /* change the volume of the waveform data */
    lp      = lpWaveData;
    iFactor = (bIncrease ? 125 : 75);
    for (dw = 0; dw < dwWaveDataSize; dw++)
    {
        iTmp = (((int)(BYTE)(*lp) - MIDVALUE) * iFactor) / 100;
        *lp = (BYTE)max( MINVALUE, min(iTmp+MIDVALUE, MAXVALUE) );
        lp++;
    }

    /* set file dirty flag */
    bIsFileDirty = TRUE;

    /* paint the new waveform data */
    PaintWaveData( hDlg );

    return;
}

/*--------------------------------------------------------------------
        PLAY WAVE FILE
        Play waveform data
  --------------------------------------------------------------------*/

static void PlayWaveFile ( HWND hDlg )
{
    LPSTR lpsz;
    DWORD dwFrom;
    int   iMinPos, iMaxPos;
    HWND  hwndScrollBar = GetDlgItem( hDlg, IDD_SCROLLBAR );

    /* anything to play? */
    if (!lpWaveData)
    {
        return;
    }

    /* get waveform file to play */
    if (!bIsFileDirty)
    {
        lpsz = szFileName;
    }
    else
    {
        /* temporarily store waveform data to disk for playing */
        WriteWaveData( hDlg, szTempFileName, lpWaveData,
                       dwWaveDataSize, dwSamplesPerSec );
        lpsz = szTempFileName;
    }

    /* if current position is end of sound, reset to beginning */
    ScrollBar_GetRange( hwndScrollBar, &iMinPos, &iMaxPos );
    if (iHScrollPos == iMaxPos)
    {
        FORWARD_WM_HSCROLL( hDlg, hwndScrollBar, 0, SB_TOP, SendMessage );
    }

    /* convert current sample position to milliseconds */
    dwFrom = (dwCurrentSample*1000) / dwSamplesPerSec;

    /* play waveform file */
    if (OpenDevice( hDlg, lpsz, &mciDevice ))
    {
        if (!BeginPlay( hDlg, mciDevice, dwFrom))
        {
            CloseAudioDevice( hDlg );
            return;
        }

        /* set timer to update display */
        uTimerID = TMRPLAY;
        SetTimer( hDlg, uTimerID, 100, NULL );
    }
    return;
}

/*--------------------------------------------------------------------
        RECORD WAVE FILE
        Record waveform data
  --------------------------------------------------------------------*/

static void RecordWaveFile ( HWND hDlg )
{
    HINSTANCE hInstance;
    char szErrStr[MAX_RSRC_STRING_LEN];
    char szCaption[MAX_RSRC_STRING_LEN];

    /* for demo purposes, record only onto new wave files */
    if (lpWaveData)
    {
        hInstance = GetWindowInstance( hDlg );
        LoadString( hInstance, IDS_BADRECORDFILE, szErrStr,
                    sizeof(szErrStr) );
        LoadString( hInstance, IDS_CAPTION, szCaption,
                    sizeof(szCaption) );
        MessageBox( hDlg, szErrStr, szCaption,
                    MB_ICONEXCLAMATION | MB_OK );
        return;
    }

    /* record waveform data into a new file */
    if (OpenDevice( hDlg, "", &mciDevice))
    {
        /* set recording to stop after 20 seconds */
        if (!BeginRecord( hDlg, mciDevice, 20000))
        {
            CloseAudioDevice( hDlg );
            return;
        }

        /* set timer to update display */
        uTimerID = TMRRECORD;
        SetTimer( hDlg, uTimerID, 100, NULL );
    }
    return;
}
